@@ -1,8 +1,8 @@
 //! The solana-program-test provides a BanksClient-based test framework BPF programs
 #![allow(clippy::integer_arithmetic)]
 
-// Export types so test clients can limit their solana crate dependencies
-pub use solana_banks_client::BanksClient;
+#[allow(deprecated)]
+use solana_sdk::sysvar::fees::Fees;
 // Export tokio for test clients
 pub use tokio;
 use {
@@ -11,9 +11,11 @@ use {
     log::*,
     solana_banks_client::start_client,
     solana_banks_server::banks_server::start_local_server,
+    solana_program_runtime::{ic_msg, invoke_context::ProcessInstructionWithContext, stable_log},
     solana_runtime::{
-        bank::{Bank, Builtin},
+        bank::{Bank, ExecuteTimings},
         bank_forks::BankForks,
+        builtins::Builtin,
         commitment::BlockCommitmentCache,
         genesis_utils::{create_genesis_config_with_leader_ex, GenesisConfigInfo},
     },
@@ -21,9 +23,9 @@ use {
         account::{Account, AccountSharedData, ReadableAccount, WritableAccount},
         account_info::AccountInfo,
         clock::{Clock, Slot},
+        compute_budget::ComputeBudget,
         entrypoint::{ProgramResult, SUCCESS},
         epoch_schedule::EpochSchedule,
-        execute_timings::ExecuteTimings,
         feature_set::demote_program_write_locks,
         fee_calculator::{FeeCalculator, FeeRateGovernor},
         genesis_config::{ClusterType, GenesisConfig},
@@ -32,16 +34,13 @@ use {
         message::Message,
         native_token::sol_to_lamports,
         poh_config::PohConfig,
-        process_instruction::{
-            self, stable_log, BpfComputeBudget, InvokeContext, ProcessInstructionWithContext,
-        },
         program_error::{ProgramError, ACCOUNT_BORROW_FAILED, UNSUPPORTED_SYSVAR},
         pubkey::Pubkey,
         rent::Rent,
         signature::{Keypair, Signer},
         sysvar::{
             clock, epoch_schedule,
-            fees::{self, Fees},
+            fees::{self},
             rent, Sysvar,
         },
     },
@@ -64,6 +63,8 @@ use {
     thiserror::Error,
     tokio::task::JoinHandle,
 };
+// Export types so test clients can limit their solana crate dependencies
+pub use {solana_banks_client::BanksClient, solana_program_runtime::invoke_context::InvokeContext};
 
 pub mod programs;
 
@@ -79,33 +80,34 @@ pub enum ProgramTestError {
 }
 
 thread_local! {
-    static INVOKE_CONTEXT: RefCell<Option<(usize, usize)>> = RefCell::new(None);
+    static INVOKE_CONTEXT: RefCell<Option<usize>> = RefCell::new(None);
 }
-fn set_invoke_context(new: &mut dyn InvokeContext) {
-    INVOKE_CONTEXT.with(|invoke_context| unsafe {
-        invoke_context.replace(Some(transmute::<_, (usize, usize)>(new)))
-    });
+fn set_invoke_context(new: &mut InvokeContext) {
+    INVOKE_CONTEXT
+        .with(|invoke_context| unsafe { invoke_context.replace(Some(transmute::<_, usize>(new))) });
 }
-fn get_invoke_context<'a>() -> &'a mut dyn InvokeContext {
-    let fat = INVOKE_CONTEXT.with(|invoke_context| match *invoke_context.borrow() {
+fn get_invoke_context<'a, 'b>() -> &'a mut InvokeContext<'b> {
+    let ptr = INVOKE_CONTEXT.with(|invoke_context| match *invoke_context.borrow() {
         Some(val) => val,
         None => panic!("Invoke context not set!"),
     });
-    unsafe { transmute::<(usize, usize), &mut dyn InvokeContext>(fat) }
+    unsafe { transmute::<usize, &mut InvokeContext>(ptr) }
 }
 
 pub fn builtin_process_instruction(
     process_instruction: solana_sdk::entrypoint::ProcessInstruction,
-    program_id: &Pubkey,
+    _first_instruction_account: usize,
     input: &[u8],
-    invoke_context: &mut dyn InvokeContext,
+    invoke_context: &mut InvokeContext,
 ) -> Result<(), InstructionError> {
     set_invoke_context(invoke_context);
 
-    let logger = invoke_context.get_logger();
-    stable_log::program_invoke(&logger, program_id, invoke_context.invoke_depth());
+    let log_collector = invoke_context.get_log_collector();
+    let program_id = invoke_context.get_caller()?;
+    stable_log::program_invoke(&log_collector, program_id, invoke_context.invoke_depth());
 
-    let keyed_accounts = invoke_context.get_keyed_accounts()?;
+    // Skip the processor account
+    let keyed_accounts = &invoke_context.get_keyed_accounts()?[1..];
 
     // Copy all the accounts into a HashMap to ensure there are no duplicates
     let mut accounts: HashMap<Pubkey, Account> = keyed_accounts
@@ -155,10 +157,10 @@ pub fn builtin_process_instruction(
     // Execute the program
     process_instruction(program_id, &account_infos, input).map_err(|err| {
         let err = u64::from(err);
-        stable_log::program_failure(&logger, program_id, &err.into());
+        stable_log::program_failure(&log_collector, program_id, &err.into());
         err
     })?;
-    stable_log::program_success(&logger, program_id);
+    stable_log::program_success(&log_collector, program_id);
 
     // Commit AccountInfo changes back into KeyedAccounts
     for keyed_account in keyed_accounts {
@@ -178,12 +180,12 @@ pub fn builtin_process_instruction(
 macro_rules! processor {
     ($process_instruction:expr) => {
         Some(
-            |program_id: &Pubkey,
+            |first_instruction_account: usize,
              input: &[u8],
-             invoke_context: &mut dyn solana_sdk::process_instruction::InvokeContext| {
+             invoke_context: &mut solana_program_test::InvokeContext| {
                 $crate::builtin_process_instruction(
                     $process_instruction,
-                    program_id,
+                    first_instruction_account,
                     input,
                     invoke_context,
                 )
@@ -202,12 +204,13 @@ fn get_sysvar<T: Default + Sysvar + Sized + serde::de::DeserializeOwned>(
         .try_borrow_mut()
         .map_err(|_| ACCOUNT_BORROW_FAILED)
         .unwrap()
-        .consume(invoke_context.get_bpf_compute_budget().sysvar_base_cost + T::size_of() as u64)
+        .consume(invoke_context.get_compute_budget().sysvar_base_cost + T::size_of() as u64)
         .is_err()
     {
         panic!("Exceeded compute budget");
     }
-    match process_instruction::get_sysvar::<T>(invoke_context, id) {
+
+    match invoke_context.get_sysvar::<T>(id) {
         Ok(sysvar_data) => unsafe {
             *(var_addr as *mut _ as *mut T) = sysvar_data;
             SUCCESS
@@ -220,11 +223,7 @@ struct SyscallStubs {}
 impl solana_sdk::program_stubs::SyscallStubs for SyscallStubs {
     fn sol_log(&self, message: &str) {
         let invoke_context = get_invoke_context();
-        let logger = invoke_context.get_logger();
-        let logger = logger.borrow_mut();
-        if logger.log_enabled() {
-            logger.log(&format!("Program log: {}", message));
-        }
+        ic_msg!(invoke_context, "Program log: {}", message);
     }
 
     fn sol_invoke_signed(
@@ -239,22 +238,15 @@ impl solana_sdk::program_stubs::SyscallStubs for SyscallStubs {
         //
 
         let invoke_context = get_invoke_context();
-        let logger = invoke_context.get_logger();
+        let log_collector = invoke_context.get_log_collector();
 
         let caller = *invoke_context.get_caller().expect("get_caller");
         let message = Message::new(&[instruction.clone()], None);
         let program_id_index = message.instructions[0].program_id_index as usize;
         let program_id = message.account_keys[program_id_index];
-        let program_account_info = || {
-            for account_info in account_infos {
-                if account_info.unsigned_key() == &program_id {
-                    return account_info;
-                }
-            }
-            panic!("Program id {} wasn't found in account_infos", program_id);
-        };
-        let demote_program_write_locks =
-            invoke_context.is_feature_active(&demote_program_write_locks::id());
+        let demote_program_write_locks = invoke_context
+            .feature_set
+            .is_active(&demote_program_write_locks::id());
         // TODO don't have the caller's keyed_accounts so can't validate writer or signer escalation or deescalation yet
         let caller_privileges = message
             .account_keys
@@ -263,38 +255,40 @@ impl solana_sdk::program_stubs::SyscallStubs for SyscallStubs {
             .map(|(i, _)| message.is_writable(i, demote_program_write_locks))
             .collect::<Vec<bool>>();
 
-        stable_log::program_invoke(&logger, &program_id, invoke_context.invoke_depth());
-
-        fn ai_to_a(ai: &AccountInfo) -> AccountSharedData {
-            AccountSharedData::from(Account {
-                lamports: ai.lamports(),
-                data: ai.try_borrow_data().unwrap().to_vec(),
-                owner: *ai.owner,
-                executable: ai.executable,
-                rent_epoch: ai.rent_epoch,
-            })
-        }
-        let executables = vec![(
-            program_id,
-            Rc::new(RefCell::new(ai_to_a(program_account_info()))),
-        )];
+        stable_log::program_invoke(&log_collector, &program_id, invoke_context.invoke_depth());
 
         // Convert AccountInfos into Accounts
-        let mut accounts = vec![];
-        'outer: for key in &message.account_keys {
-            for account_info in account_infos {
-                if account_info.unsigned_key() == key {
-                    accounts.push((*key, Rc::new(RefCell::new(ai_to_a(account_info)))));
-                    continue 'outer;
-                }
+        let mut account_indices = Vec::with_capacity(message.account_keys.len());
+        let mut accounts = Vec::with_capacity(message.account_keys.len());
+        for (i, account_key) in message.account_keys.iter().enumerate() {
+            let ((account_index, account), account_info) = invoke_context
+                .get_account(account_key)
+                .zip(
+                    account_infos
+                        .iter()
+                        .find(|account_info| account_info.unsigned_key() == account_key),
+                )
+                .ok_or(InstructionError::MissingAccount)
+                .unwrap();
+            {
+                let mut account = account.borrow_mut();
+                account.copy_into_owner_from_slice(account_info.owner.as_ref());
+                account.set_data_from_slice(&account_info.try_borrow_data().unwrap());
+                account.set_lamports(account_info.lamports());
+                account.set_executable(account_info.executable);
+                account.set_rent_epoch(account_info.rent_epoch);
             }
-            panic!("Account {} wasn't found in account_infos", key);
+            let account_info = if message.is_writable(i, demote_program_write_locks) {
+                Some(account_info)
+            } else {
+                None
+            };
+            account_indices.push(account_index);
+            accounts.push((account, account_info));
         }
-        assert_eq!(
-            accounts.len(),
-            message.account_keys.len(),
-            "Missing or not enough accounts passed to invoke"
-        );
+        let (program_account_index, _program_account) =
+            invoke_context.get_account(&program_id).unwrap();
+        let program_indices = vec![program_account_index];
 
         // Check Signers
         for account_info in account_infos {
@@ -311,59 +305,57 @@ impl solana_sdk::program_stubs::SyscallStubs for SyscallStubs {
                             break;
                         }
                     }
-                    if !program_signer {
-                        panic!("Missing signer for {}", instruction_account.pubkey);
-                    }
+                    assert!(
+                        program_signer,
+                        "Missing signer for {}",
+                        instruction_account.pubkey
+                    );
                 }
             }
         }
 
-        invoke_context.record_instruction(instruction);
+        if let Some(instruction_recorder) = &invoke_context.instruction_recorder {
+            instruction_recorder.record_instruction(instruction.clone());
+        }
 
-        solana_runtime::message_processor::MessageProcessor::process_cross_program_instruction(
-            &message,
-            &executables,
-            &accounts,
-            &caller_privileges,
-            invoke_context,
-        )
-        .map_err(|err| ProgramError::try_from(err).unwrap_or_else(|err| panic!("{}", err)))?;
+        invoke_context
+            .process_instruction(
+                &message,
+                &message.instructions[0],
+                &program_indices,
+                &account_indices,
+                &caller_privileges,
+            )
+            .map_err(|err| ProgramError::try_from(err).unwrap_or_else(|err| panic!("{}", err)))?;
 
         // Copy writeable account modifications back into the caller's AccountInfos
-        for (i, (pubkey, account)) in accounts.iter().enumerate().take(message.account_keys.len()) {
-            if !message.is_writable(i, demote_program_write_locks) {
-                continue;
-            }
-            for account_info in account_infos {
-                if account_info.unsigned_key() == pubkey {
-                    **account_info.try_borrow_mut_lamports().unwrap() = account.borrow().lamports();
-
-                    let mut data = account_info.try_borrow_mut_data()?;
-                    let account_borrow = account.borrow();
-                    let new_data = account_borrow.data();
-                    if account_info.owner != account.borrow().owner() {
-                        // TODO Figure out a better way to allow the System Program to set the account owner
-                        #[allow(clippy::transmute_ptr_to_ptr)]
-                        #[allow(mutable_transmutes)]
-                        let account_info_mut =
-                            unsafe { transmute::<&Pubkey, &mut Pubkey>(account_info.owner) };
-                        *account_info_mut = *account.borrow().owner();
-                    }
-                    if data.len() != new_data.len() {
-                        // TODO: Figure out how to allow the System Program to resize the account data
-                        panic!(
-                            "Account data resizing not supported yet: {} -> {}. \
-                            Consider making this test conditional on `#[cfg(feature = \"test-bpf\")]`",
-                            data.len(),
-                            new_data.len()
-                        );
-                    }
-                    data.clone_from_slice(new_data);
+        for (account, account_info) in accounts.iter() {
+            if let Some(account_info) = account_info {
+                **account_info.try_borrow_mut_lamports().unwrap() = account.borrow().lamports();
+                let mut data = account_info.try_borrow_mut_data()?;
+                let account_borrow = account.borrow();
+                let new_data = account_borrow.data();
+                if account_info.owner != account.borrow().owner() {
+                    // TODO Figure out a better way to allow the System Program to set the account owner
+                    #[allow(clippy::transmute_ptr_to_ptr)]
+                    #[allow(mutable_transmutes)]
+                    let account_info_mut =
+                        unsafe { transmute::<&Pubkey, &mut Pubkey>(account_info.owner) };
+                    *account_info_mut = *account.borrow().owner();
                 }
+                // TODO: Figure out how to allow the System Program to resize the account data
+                assert!(
+                    data.len() == new_data.len(),
+                    "Account data resizing not supported yet: {} -> {}. \
+                        Consider making this test conditional on `#[cfg(feature = \"test-bpf\")]`",
+                    data.len(),
+                    new_data.len()
+                );
+                data.clone_from_slice(new_data);
             }
         }
 
-        stable_log::program_success(&logger, &program_id);
+        stable_log::program_success(&log_collector, &program_id);
         Ok(())
     }
 
@@ -375,6 +367,7 @@ impl solana_sdk::program_stubs::SyscallStubs for SyscallStubs {
         get_sysvar::<EpochSchedule>(&epoch_schedule::id(), var_addr)
     }
 
+    #[allow(deprecated)]
     fn sol_get_fees_sysvar(&self, var_addr: *mut u8) -> u64 {
         get_sysvar::<Fees>(&fees::id(), var_addr)
     }
@@ -418,9 +411,9 @@ pub fn read_file<P: AsRef<Path>>(path: P) -> Vec<u8> {
     file_data
 }
 
-fn setup_fee_calculator(bank: Bank) -> Bank {
-    // Realistic fee_calculator part 1: Fake a single signature by calling
-    // `bank.commit_transactions()` so that the fee calculator in the child bank will be
+fn setup_fees(bank: Bank) -> Bank {
+    // Realistic fees part 1: Fake a single signature by calling
+    // `bank.commit_transactions()` so that the fee in the child bank will be
     // initialized with a non-zero fee.
     assert_eq!(bank.signature_count(), 0);
     bank.commit_transactions(
@@ -438,20 +431,16 @@ fn setup_fee_calculator(bank: Bank) -> Bank {
     let bank = Bank::new_from_parent(&bank, bank.collector_id(), bank.slot() + 1);
     debug!("Bank slot: {}", bank.slot());
 
-    // Realistic fee_calculator part 2: Tick until a new blockhash is produced to pick up the
-    // non-zero fee calculator
+    // Realistic fees part 2: Tick until a new blockhash is produced to pick up the
+    // non-zero fees
     let last_blockhash = bank.last_blockhash();
     while last_blockhash == bank.last_blockhash() {
         bank.register_tick(&Hash::new_unique());
     }
-    let last_blockhash = bank.last_blockhash();
-    // Make sure the new last_blockhash now requires a fee
-    assert_ne!(
-        bank.get_fee_calculator(&last_blockhash)
-            .expect("fee_calculator")
-            .lamports_per_signature,
-        0
-    );
+
+    // Make sure a fee is now required
+    let lamports_per_signature = bank.get_lamports_per_signature();
+    assert_ne!(lamports_per_signature, 0);
 
     bank
 }
@@ -459,7 +448,7 @@ fn setup_fee_calculator(bank: Bank) -> Bank {
 pub struct ProgramTest {
     accounts: Vec<(Pubkey, AccountSharedData)>,
     builtins: Vec<Builtin>,
-    bpf_compute_max_units: Option<u64>,
+    compute_max_units: Option<u64>,
     prefer_bpf: bool,
     use_bpf_jit: bool,
 }
@@ -489,7 +478,7 @@ impl Default for ProgramTest {
         Self {
             accounts: vec![],
             builtins: vec![],
-            bpf_compute_max_units: None,
+            compute_max_units: None,
             prefer_bpf,
             use_bpf_jit: false,
         }
@@ -519,9 +508,16 @@ impl ProgramTest {
         self.prefer_bpf = prefer_bpf;
     }
 
+    /// Override the default maximum compute units
+    pub fn set_compute_max_units(&mut self, compute_max_units: u64) {
+        self.compute_max_units = Some(compute_max_units);
+    }
+
     /// Override the BPF compute budget
+    #[allow(deprecated)]
+    #[deprecated(since = "1.8.0", note = "please use `set_compute_max_units` instead")]
     pub fn set_bpf_compute_max_units(&mut self, bpf_compute_max_units: u64) {
-        self.bpf_compute_max_units = Some(bpf_compute_max_units);
+        self.compute_max_units = Some(bpf_compute_max_units);
     }
 
     /// Execute the BPF program with JIT if true, interpreted if false
@@ -758,12 +754,12 @@ impl ProgramTest {
         debug!("Payer address: {}", mint_keypair.pubkey());
         debug!("Genesis config: {}", genesis_config);
 
-        let mut bank = Bank::new(&genesis_config);
+        let mut bank = Bank::new_for_tests(&genesis_config);
 
         // Add loaders
         macro_rules! add_builtin {
             ($b:expr) => {
-                bank.add_builtin(&$b.0, $b.1, $b.2)
+                bank.add_builtin(&$b.0, &$b.1, $b.2)
             };
         }
         add_builtin!(solana_bpf_loader_deprecated_program!());
@@ -784,7 +780,7 @@ impl ProgramTest {
         for builtin in self.builtins.iter() {
             bank.add_builtin(
                 &builtin.name,
-                builtin.id,
+                &builtin.id,
                 builtin.process_instruction_with_context,
             );
         }
@@ -796,13 +792,13 @@ impl ProgramTest {
             bank.store_account(address, account);
         }
         bank.set_capitalization();
-        if let Some(max_units) = self.bpf_compute_max_units {
-            bank.set_bpf_compute_budget(Some(BpfComputeBudget {
+        if let Some(max_units) = self.compute_max_units {
+            bank.set_compute_budget(Some(ComputeBudget {
                 max_units,
-                ..BpfComputeBudget::default()
+                ..ComputeBudget::default()
             }));
         }
-        let bank = setup_fee_calculator(bank);
+        let bank = setup_fees(bank);
         let slot = bank.slot();
         let last_blockhash = bank.last_blockhash();
         let bank_forks = Arc::new(RwLock::new(BankForks::new(bank)));
@@ -818,7 +814,6 @@ impl ProgramTest {
                 genesis_config,
                 mint_keypair,
                 voting_keypair,
-                validator_pubkey: bootstrap_validator_pubkey,
             },
         )
     }
@@ -882,18 +877,25 @@ impl ProgramTest {
 
 #[async_trait]
 pub trait ProgramTestBanksClientExt {
+    /// Get a new blockhash, similar in spirit to RpcClient::get_new_blockhash()
+    ///
+    /// This probably should eventually be moved into BanksClient proper in some form
+    #[deprecated(
+        since = "1.9.0",
+        note = "Please use `get_new_latest_blockhash `instead"
+    )]
     async fn get_new_blockhash(&mut self, blockhash: &Hash) -> io::Result<(Hash, FeeCalculator)>;
+    /// Get a new latest blockhash, similar in spirit to RpcClient::get_latest_blockhash()
+    async fn get_new_latest_blockhash(&mut self, blockhash: &Hash) -> io::Result<Hash>;
 }
 
 #[async_trait]
 impl ProgramTestBanksClientExt for BanksClient {
-    /// Get a new blockhash, similar in spirit to RpcClient::get_new_blockhash()
-    ///
-    /// This probably should eventually be moved into BanksClient proper in some form
     async fn get_new_blockhash(&mut self, blockhash: &Hash) -> io::Result<(Hash, FeeCalculator)> {
         let mut num_retries = 0;
         let start = Instant::now();
         while start.elapsed().as_secs() < 5 {
+            #[allow(deprecated)]
             if let Ok((fee_calculator, new_blockhash, _slot)) = self.get_fees().await {
                 if new_blockhash != *blockhash {
                     return Ok((new_blockhash, fee_calculator));
@@ -915,6 +917,31 @@ impl ProgramTestBanksClientExt for BanksClient {
             ),
         ))
     }
+
+    async fn get_new_latest_blockhash(&mut self, blockhash: &Hash) -> io::Result<Hash> {
+        let mut num_retries = 0;
+        let start = Instant::now();
+        while start.elapsed().as_secs() < 5 {
+            let new_blockhash = self.get_latest_blockhash().await?;
+            if new_blockhash != *blockhash {
+                return Ok(new_blockhash);
+            }
+            debug!("Got same blockhash ({:?}), will retry...", blockhash);
+
+            tokio::time::sleep(Duration::from_millis(200)).await;
+            num_retries += 1;
+        }
+
+        Err(io::Error::new(
+            io::ErrorKind::Other,
+            format!(
+                "Unable to get new blockhash after {}ms (retried {} times), stuck at {}",
+                start.elapsed().as_millis(),
+                num_retries,
+                blockhash
+            ),
+        ))
+    }
 }
 
 struct DroppableTask<T>(Arc<AtomicBool>, JoinHandle<T>);
